(*
 * F-ing Modules with Coq
 *
 * (c) 2009 Andreas Rossberg & Claudio Russo
 *)

Require Import Metatheory.
Require Import AdditionalTactics.
Require Import Misc.
Require Import Fomega.


(*************************************************************************)
(* Lists of variable terms *)
(*************************************************************************)

Definition freeTs (as' : list var) : list typ := map varT (map free as').
Definition freeEs (as' : list var) : list exp := map varE (map free as').


Lemma freeTs_cons : forall (a : var) as',
  freeTs (a::as') = varT a :: freeTs as'.
Proof.
  intros. unfold freeTs, map. reflexivity.
Qed.

Lemma freeEs_cons : forall (x : var) xs,
  freeEs (x::xs) = varE x :: freeEs xs.
Proof.
  intros. unfold freeEs, map. reflexivity.
Qed.


Lemma freeTs_concat : forall as1 as2,
  freeTs (as1 ++ as2) = freeTs as1 ++ freeTs as2.
Proof.
  induction as1; intro as2; simpl.
    reflexivity.
    unfold freeTs in *. do 2 simpl map. f_equal. apply IHas1.
Qed.

Lemma freeEs_concat : forall xs1 xs2,
  freeEs (xs1 ++ xs2) = freeEs xs1 ++ freeEs xs2.
Proof.
  induction xs1; intro xs2; simpl.
    reflexivity.
    unfold freeEs in *. do 2 simpl map. f_equal. apply IHxs1.
Qed.


Lemma freeTs_length : forall as',
  length (freeTs as') = length as'.
Proof.
  intros as'. unfold freeTs. rewrite map_length. rewrite map_length. reflexivity.
Qed.

Lemma freeEs_length : forall xs,
  length (freeEs xs) = length xs.
Proof.
  intros xs. unfold freeEs. rewrite map_length. rewrite map_length. reflexivity.
Qed.


Lemma lclosedT_vars : forall as' t,
  In t (freeTs as') ->
  lclosedT t.
Proof.
  unfold freeTs.
  induction as'; simpl; intros t Elem.
    inversion Elem.
    destruct Elem; subst.
      apply lclosedT_var.
      apply IHas'. apply H.
Qed.

Lemma lclosedE_vars : forall xs e,
  In e (freeEs xs) ->
  lclosedE e.
Proof.
  unfold freeEs.
  induction xs; simpl; intros e Elem.
    inversion Elem.
    destruct Elem; subst.
      apply lclosedE_var.
      apply IHxs. apply H.
Qed.


(*************************************************************************)
(* Lists of bound variables as terms *)
(*************************************************************************)

Fixpoint enum n : list nat :=
  match n with
  | O => nil
  | S n' => n' :: enum n'
 end.

Definition bnds n : list occ := map bnd (enum n).
Definition bndTs n : list typ := map varT (bnds n).
Definition bndEs n : list exp := map varE (bnds n).


Lemma bnd_upper : forall n m i,
  In i (map (fun j => n + j) (enum m)) ->
  n <= i.
Proof.
  intro n. induction m; intros i' Elem'; simpl enum in *; simpl map in *.
    inversion Elem'.
    destruct Elem'. omega. apply IHm. apply H.
Qed.

Lemma bndTs_length : forall n,
  length (bndTs n) = n.
Proof.
  unfold bndTs.
  induction n; simpl fold_right; simpl length.
    reflexivity.
    f_equal. apply IHn.
Qed.

Lemma bndEs_length : forall n,
  length (bndEs n) = n.
Proof.
  unfold bndEs.
  induction n; simpl fold_right; simpl length.
    reflexivity.
    f_equal. apply IHn.
Qed.


Lemma substAT_free_bndTs : forall (a : var) t n,
  map (substAT a t) (bndTs n) = bndTs n.
Proof.
  intros a t n.
  induction n; unfold bndTs.
    reflexivity.
    unfold bnds. simpl enum. do 2 simpl map.
    unfold bndTs in IHn. unfold bnds in IHn. rewrite IHn. reflexivity.
Qed.

Lemma substAT_bnd_bndTs : forall (n : nat) t m,
  m <= n ->
  map (substAT n t) (bndTs m) = bndTs m.
Proof.
  intros n t m H.
  induction m; unfold bndTs; simpl fold_right; simpl map.
    reflexivity.
    simpl length in H. unfold bndTs in IHm. unfold bnds in IHm. rewrite IHm.
      destruct (m === n).
        absurd (S m <= m). omega. omega.
        reflexivity.
      omega.
Qed.


(*************************************************************************)
(* Lists of local closure propositions *)
(*************************************************************************)

Definition lclosedTs ts := forall t, In t ts -> lclosedT t.
Definition lclosedEs es := forall e, In e es -> lclosedE e.

Lemma lclosedTs_head : forall t ts,
  lclosedTs (t::ts) ->
  lclosedT t.
Proof.
  intros t ts H. apply H. apply in_eq.
Qed.

Lemma lclosedTs_tail : forall t ts,
  lclosedTs (t::ts) ->
  lclosedTs ts.
Proof.
  intros t ts H t' Elem. apply H. apply in_cons. apply Elem.
Qed.

Lemma lclosedTs_rest : forall t t' ts,
  lclosedTs (t'::ts) ->
  In t ts ->
  lclosedT t.
Proof.
  intros t t' ts H Elem. apply H. apply in_cons. apply Elem.
Qed.


Lemma lclosedTs_nil :
  lclosedTs nil.
Proof.
  intros t Elem. inversion Elem.
Qed.

Lemma lclosedTs_cons : forall t ts,
  lclosedT t ->
  lclosedTs ts ->
  lclosedTs (t::ts).
Proof.
  intros t ts Ht Hts t' Elem.
  destruct ts; simpl In in Elem; destruct Elem; subst.
    apply Ht.
    destruct H.
    apply Ht.
    destruct H; subst.
      eapply lclosedTs_head. apply Hts.
      eapply lclosedTs_rest. apply Hts. apply H.
Qed.

Lemma lclosedTs_concat : forall ts1 ts2,
  lclosedTs ts1 ->
  lclosedTs ts2 ->
  lclosedTs (ts1 ++ ts2).
Proof.
  induction ts1; intros ts2 Hts1 Hts2; simpl.
    apply Hts2.
    apply lclosedTs_cons.
      eapply lclosedTs_head. apply Hts1.
      apply IHts1. eapply lclosedTs_tail. apply Hts1. apply Hts2.
Qed.

Lemma lclosedEs_head : forall e es,
  lclosedEs (e::es) ->
  lclosedE e.
Proof.
  intros e es H. apply H. apply in_eq.
Qed.

Lemma lclosedEs_tail : forall e es,
  lclosedEs (e::es) ->
  lclosedEs es.
Proof.
  intros e es H e' Elem. apply H. apply in_cons. apply Elem.
Qed.

Lemma lclosedEs_rest : forall e e' es,
  lclosedEs (e'::es) ->
  In e es ->
  lclosedE e.
Proof.
  intros e e' es H Elem. apply H. apply in_cons. apply Elem.
Qed.


Lemma lclosedEs_nil :
  lclosedEs nil.
Proof.
  intros t Elem. inversion Elem.
Qed.

Lemma lclosedEs_cons : forall e es,
  lclosedE e ->
  lclosedEs es ->
  lclosedEs (e::es).
Proof.
  intros e es He Hes e' Elem.
  destruct es; simpl In in Elem; destruct Elem; subst.
    apply He.
    destruct H.
    apply He.
    destruct H; subst.
      eapply lclosedEs_head. apply Hes.
      eapply lclosedEs_rest. apply Hes. apply H.
Qed.


Lemma lclosedTs_vars : forall as',
  lclosedTs (freeTs as').
Proof.
  intros as' t Elem. eapply lclosedT_vars. apply Elem.
Qed.

Lemma lclosedEs_vars : forall xs,
  lclosedEs (freeEs xs).
Proof.
  intros xs e Elem. eapply lclosedE_vars. apply Elem.
Qed.


Lemma regularityTs : forall C ts ks,
  length ks = length ts ->
  (forall t k, In (t, k) (combine ts ks) -> ofT C t k) ->
  lclosedTs ts.
Proof.
  intro C.
  induction ts; intros ks Len Hts; destruct ks; inversion Len.
    apply lclosedTs_nil.
    apply lclosedTs_cons.
      eapply regularityT. apply Hts. simpl combine. apply in_eq.
      eapply IHts. apply H0. intros t' k' Elem. apply Hts. simpl combine. apply in_cons. apply Elem.
Qed.

Lemma regularityEs : forall C G es ts,
  length ts = length es ->
  (forall e t, In (e, t) (combine es ts) -> ofE C G e t) ->
  lclosedEs es.
Proof.
  intros C G.
  induction es; intros ts Len Hes; destruct ts; inversion Len.
    apply lclosedEs_nil.
    apply lclosedEs_cons.
      eapply regularityE. apply Hes. simpl combine. apply in_eq.
      eapply IHes. apply H0. intros e' t' Elem. apply Hes. simpl combine. apply in_cons. apply Elem.
Qed.


(*************************************************************************)
(* Shifting variables *)
(*************************************************************************)

Fixpoint shifts j (x : occ) {struct j} : occ :=
  match j with
  | O => x
  | S j' => shift (shifts j' x)
  end.

Lemma shifts_free : forall j (v : var),
  shifts j v = v.
Proof.
  induction j; intros v; simpl shifts.
    Case "j = 0".
      f_equal.
    Case "j = S j'".
      rewrite IHj. simpl shift. f_equal.
Qed.

Lemma shifts_bnd : forall j (n : nat),
  shifts j n = j + n.
Proof.
  induction j; intros n; simpl shifts.
    Case "j = 0".
      f_equal.
    Case "j = S j'".
      rewrite IHj. simpl shift. f_equal.
Qed.

Lemma shift_shifts : forall j x,
  shift (shifts j x) = shifts (S j) x.
Proof.
  intros j x. reflexivity.
Qed.

Lemma shifts_shift : forall j x,
  shifts j (shift x) = shifts (S j) x.
Proof.
  induction j; intros x; simpl shifts.
    Case "j = 0".
      reflexivity.
    Case "j = S j'".
      rewrite IHj. rewrite <- shift_shifts. reflexivity.
Qed.

Lemma shifts_shifts : forall j1 j2 x,
  shifts j1 (shifts j2 x) = shifts (j1 + j2) x.
Proof.
  induction j1; intros j2 x; simpl shifts.
    Case "j1 = 0".
      reflexivity.
    Case "j1 = S j1'".
      rewrite IHj1. rewrite shift_shifts. reflexivity.
Qed.


Lemma shifts_injectivity : forall i a1 a2,
  shifts i a1 = shifts i a2 ->
  a1 = a2.
Proof.
  unfold shifts.
  induction i; intros a1 a2.
  Case "0".
    intros Eq. apply Eq.
  Case "S i'".
    intros Eq.
    apply IHi. apply shift_injectivity. apply Eq.
Qed.


Lemma map_shifts_bnds : forall i n,
  map (shifts i) (bnds n) = map bnd (map (fun j => i + j) (enum n)).
Proof.
  induction n; unfold bnds; simpl enum; do 2 simpl map.
    reflexivity.
    fold (bnds n). rewrite IHn. rewrite shifts_bnd. reflexivity.
Qed.


(*************************************************************************)
(* Multiple substitution *)
(*************************************************************************)

Fixpoint substsAT (as' : list occ) (ts : list typ) t {struct as'} : typ :=
  match as' with
  | nil => t
  | a::as'' =>
    match ts with
    | nil => t
    | t'::ts' => substAT a t' (substsAT as'' ts' t)
    end
  end.

Fixpoint substsAE (as' : list occ) (ts : list typ) e {struct as'} : exp :=
  match as' with
  | nil => e
  | a::as'' =>
    match ts with
    | nil => e
    | t'::ts' => substAE a t' (substsAE as'' ts' e)
    end
  end.

Fixpoint substsXE (xs : list occ) (es : list exp) e {struct xs} : exp :=
  match xs with
  | nil => e
  | x::xs' =>
    match es with
    | nil => e
    | e'::es' => substXE x e' (substsXE xs' es' e)
    end
  end.


(*************************************************************************)
(* Multiple opening *)
(*************************************************************************)

Fixpoint opensAT (ts : list typ) t {struct ts} : typ :=
  match ts with
  | nil => t
  | t'::ts' => substAT (length ts') t' (opensAT ts' t)
  end.

Fixpoint opensAE (ts : list typ) e {struct ts} : exp :=
  match ts with
  | nil => e
  | t::ts' => substAE (length ts') t (opensAE ts' e)
  end.

Fixpoint opensXE (es : list exp) e {struct es} : exp :=
  match es with
  | nil => e
  | e'::es' => substXE (length es') e' (opensXE es' e)
  end.


(*************************************************************************)
(* Multiple substitution vs opening *)
(*************************************************************************)

Lemma unfold_opensAT : forall ts t,
  opensAT ts t = substsAT (bnds (length ts)) ts t.
Proof.
  induction ts; intro t; simpl; try rewrite IHts; reflexivity.
Qed.

Lemma unfold_opensATs : forall ts' ts,
  map (opensAT ts') ts = map (substsAT (bnds (length ts')) ts') ts.
Proof.
  intro ts'.
  induction ts; simpl map.
    reflexivity.
    rewrite IHts. rewrite unfold_opensAT. reflexivity.
Qed.

Lemma unfold_opensAE : forall ts e,
  opensAE ts e = substsAE (bnds (length ts)) ts e.
Proof.
  induction ts; intro e; simpl; try rewrite IHts; reflexivity.
Qed.

Lemma unfold_opensXE : forall es e,
  opensXE es e = substsXE (bnds (length es)) es e.
Proof.
  induction es; intro e; simpl; try rewrite IHes; reflexivity.
Qed.


(*************************************************************************)
(* Multiple independencies *)
(*************************************************************************)

Definition indepsAT as' t := forall a, In a as' -> indepAT a t.
Definition indepsAE as' e := forall a, In a as' -> indepAE a e.
Definition indepsXE xs e := forall x, In x xs -> indepXE x e.

Definition indepATs a ts := forall t, In t ts -> indepAT a t.
Definition indepAEs a es := forall e, In e es -> indepAE a e.
Definition indepXEs x es := forall e, In e es -> indepXE x e.

Definition indepsATs as' ts := forall a, In a as' -> indepATs a ts.
Definition indepsAEs as' es := forall a, In a as' -> indepAEs a es.
Definition indepsXEs xs es := forall x, In x xs -> indepXEs x es.

Lemma indepsAT_head : forall a as' t,
  indepsAT (a::as') t ->
  indepAT a t.
Proof.
  intros a as' t H. apply H. apply in_eq.
Qed.

Lemma indepsAE_head : forall a as' e,
  indepsAE (a::as') e ->
  indepAE a e.
Proof.
  intros a as' e H. apply H. apply in_eq.
Qed.

Lemma indepsXE_head : forall x xs e,
  indepsXE (x::xs) e ->
  indepXE x e.
Proof.
  intros x xs e H. apply H. apply in_eq.
Qed.

Lemma indepsAT_tail : forall a as' t,
  indepsAT (a::as') t ->
  indepsAT as' t.
Proof.
  intros a as' t H. intros a' Elem. apply H. apply in_cons. apply Elem.
Qed.

Lemma indepsAE_tail : forall a as' e,
  indepsAE (a::as') e ->
  indepsAE as' e.
Proof.
  intros a as' e H. intros a' Elem. apply H. apply in_cons. apply Elem.
Qed.

Lemma indepsXE_tail : forall x xs e,
  indepsXE (x::xs) e ->
  indepsXE xs e.
Proof.
  intros x xs e H. intros x' Elem. apply H. apply in_cons. apply Elem.
Qed.


Lemma indepATs_head : forall a t ts,
  indepATs a (t::ts) ->
  indepAT a t.
Proof.
  intros a t ts H. apply H. apply in_eq.
Qed.

Lemma indepAEs_head : forall a e es,
  indepAEs a (e::es) ->
  indepAE a e.
Proof.
  intros a e es H. apply H. apply in_eq.
Qed.

Lemma indepXEs_head : forall x e es,
  indepXEs x (e::es) ->
  indepXE x e.
Proof.
  intros x e es H. apply H. apply in_eq.
Qed.

Lemma indepATs_tail : forall a t ts,
  indepATs a (t::ts) ->
  indepATs a ts.
Proof.
  intros a t ts H t' Elem. apply H. apply in_cons. apply Elem.
Qed.

Lemma indepAEs_tail : forall a e es,
  indepAEs a (e::es) ->
  indepAEs a es.
Proof.
  intros a e es H t' Elem. apply H. apply in_cons. apply Elem.
Qed.

Lemma indepXEs_tail : forall x e es,
  indepXEs x (e::es) ->
  indepXEs x es.
Proof.
  intros x e es H t' Elem. apply H. apply in_cons. apply Elem.
Qed.


(*************************************************************************)
(* Commuting and unfolding multiple opening *)
(*************************************************************************)

Lemma substAT_opensAT : forall i t' t ts,
  length ts <= i ->
  lclosedT t' ->
  lclosedTs ts ->
  substAT i t' (opensAT ts t) = opensAT ts (substAT i t' t).
Proof.
  intros i t' t.
  induction ts; intros Less Ht' Hts.
    Case "ts = nil".
      reflexivity.
    Case "ts = t'::ts''".
      unfold opensAT, openAT in *.
      simpl in *.
      rewrite substAT_substAT; unfold indepAT.
        rewrite IHts.
          reflexivity.
          omega.
          apply Ht'.
          eapply lclosedTs_tail. apply Hts.
        unfold not. intros. inversion H. omega.
        eapply lclosedTs_head. apply Hts.
        apply Ht'.
Qed.

Corollary opensAT_cons : forall t' ts t,
  lclosedT t' ->
  lclosedTs ts ->
  opensAT (t'::ts) t = opensAT ts (substAT (length ts) t' t).
Proof.
  intros t' ts t Ht' Hts.
  simpl.
  apply substAT_opensAT. omega. apply Ht'. apply Hts.
Qed.

Corollary opensAT_cons_vars : forall (a : var) (as' : list var) t,
  opensAT (freeTs (a::as')) t = opensAT (freeTs as') (substAT (length as') a t).
Proof.
  intros a as' t.
  unfold freeTs. simpl map.
  rewrite <- freeTs_length.
  apply opensAT_cons.
    apply lclosedT_var.
    apply lclosedTs_vars.
Qed.


Lemma substAE_opensAE : forall i t e ts,
  length ts <= i ->
  lclosedT t ->
  lclosedTs ts ->
  substAE i t (opensAE ts e) = opensAE ts (substAE i t e).
Proof.
  intros i t e.
  induction ts; intros Less Ht Hts.
    Case "ts = nil".
      reflexivity.
    Case " ts = t'::ts''".
      unfold opensAE, openAE in *.
      simpl in *.
      rewrite <- substAE_substAE; unfold indepAT.
        rewrite IHts.
          reflexivity.
          omega.
          apply Ht.
          eapply lclosedTs_tail. apply Hts.
        unfold not. intros. inversion H. omega.
        apply Ht.
        eapply lclosedTs_head. apply Hts.
Qed.

Corollary opensAE_cons : forall t ts e,
  lclosedT t ->
  lclosedTs ts ->
  opensAE (t::ts) e = opensAE ts (substAE (length ts) t e).
Proof.
  intros t ts e Ht Hts.
  simpl.
  apply substAE_opensAE. omega. apply Ht. apply Hts.
Qed.

Corollary opensAE_cons_vars : forall (a : var) (as' : list var) e,
  opensAE (freeTs (a::as')) e = opensAE (freeTs as') (substAE (length as') a e).
Proof.
  intros a as' e.
  unfold freeTs. simpl map.
  rewrite <- freeTs_length.
  apply opensAE_cons.
    apply lclosedT_var.
    apply lclosedTs_vars.
Qed.


Lemma substXE_opensXE : forall i e' e es,
  length es <= i ->
  lclosedE e' ->
  lclosedEs es ->
  substXE i e' (opensXE es e) = opensXE es (substXE i e' e).
Proof.
  intros i e' e.
  induction es; intros Less He' Hes.
    Case "es = nil".
      reflexivity.
    Case "es = e'::es''".
      unfold opensXE, openXE in *.
      simpl in *.
      rewrite <- substXE_substXE; unfold indepXE.
        rewrite IHes.
          reflexivity.
          omega.
          apply He'.
          eapply lclosedEs_tail. apply Hes.
        unfold not. intros. inversion H. omega.
        apply He'.
        eapply lclosedEs_head. apply Hes.
Qed.

Corollary opensXE_cons : forall e' es e,
  lclosedE e' ->
  lclosedEs es ->
  opensXE (e'::es) e = opensXE es (substXE (length es) e' e).
Proof.
  intros e' es t He' Hes.
  simpl.
  apply substXE_opensXE. omega. apply He'. apply Hes.
Qed.

Corollary opensXE_cons_vars : forall (x : var) (xs : list var) e,
  opensXE (freeEs (x::xs)) e = opensXE (freeEs xs) (substXE (length xs) x e).
Proof.
  intros x xs t.
  unfold freeEs. simpl map.
  rewrite <- freeEs_length.
  apply opensXE_cons.
    apply lclosedE_var.
    apply lclosedEs_vars.
Qed.


Lemma substXE_opensAE : forall ts x e' e,
  lclosedE e' ->
  substXE x e' (opensAE ts e) = opensAE ts (substXE x e' e).
Proof.
  induction ts; intros x e' e He'; simpl opensAE.
    Case "ts = nil".
      reflexivity.
    Case "ts = t::ts'".
      rewrite substXE_substAE.
        f_equal. apply IHts. apply He'.
        unfold indepAE. apply He'.
Qed.

Corollary openXE_opensAE : forall ts e' e,
  lclosedE e' ->
  openXE e' (opensAE ts e) = opensAE ts (openXE e' e).
Proof.
  intros ts e' e He'. unfold openXE. apply substXE_opensAE. apply He'.
Qed.



Lemma substAT_substsAT_bnd : forall n ts (i : nat) j t t',
  n + j <= i ->
  lclosedT t ->
  lclosedTs ts ->
  substAT i t (substsAT (map (shifts j) (bnds n)) ts t') =
  substsAT (map (shifts j) (bnds n)) ts (substAT i t t').
Proof.
  induction n; intros ts i j t t' Hi Ht Hts; unfold bnds; simpl map; simpl substsAT.
    reflexivity.
    destruct ts.
      reflexivity.
      rewrite substAT_substAT.
        rewrite IHn. reflexivity.
        omega.
        apply Ht.
        eapply lclosedTs_tail. apply Hts.
        rewrite shifts_bnd. intro Eq. inversion Eq. omega.
        simpl indepAT. eapply lclosedTs_head. apply Hts.
        rewrite shifts_bnd. simpl indepAT. apply Ht.
Qed.

Lemma opensAT_concat : forall ts1 ts2 t,
  lclosedTs ts1 ->
  lclosedTs ts2 ->
  opensAT (ts1 ++ ts2) t =
  opensAT ts2 (substsAT (map (shifts (length ts2)) (bnds (length ts1))) ts1 t).
Proof.
  induction ts1; intros ts2 t Hts1 Hts2; simpl.
    reflexivity.
    rewrite substAT_opensAT.
      rewrite IHts1. rewrite shifts_bnd. rewrite app_length.
      replace (length ts2 + length ts1) with (length ts1 + length ts2). 2: omega.
      fold (bnds (length ts1)). rewrite substAT_substsAT_bnd.
        reflexivity.
        omega.
        eapply lclosedTs_head. apply Hts1.
        eapply lclosedTs_tail. apply Hts1.
        eapply lclosedTs_tail. apply Hts1.
        apply Hts2.
        omega.
        eapply lclosedTs_head. apply Hts1.
        apply lclosedTs_concat. eapply lclosedTs_tail. apply Hts1. apply Hts2.
Qed.


Lemma substAE_substsAE_bnd : forall n ts (i : nat) j t e,
  n + j <= i ->
  lclosedT t ->
  lclosedTs ts ->
  substAE i t (substsAE (map (shifts j) (bnds n)) ts e) =
  substsAE (map (shifts j) (bnds n)) ts (substAE i t e).
Proof.
  induction n; intros ts i j t e Hi Ht Hts; unfold bnds; simpl map; simpl substsAE.
    reflexivity.
    destruct ts.
      reflexivity.
      rewrite substAE_substAE.
        rewrite IHn. reflexivity.
        omega.
        apply Ht.
        eapply lclosedTs_tail. apply Hts.
        rewrite shifts_bnd. intro Eq. inversion Eq. omega.
        simpl indepAT. eapply lclosedTs_head. apply Hts.
        rewrite shifts_bnd. simpl indepAT. apply Ht.
Qed.

Lemma opensAE_concat : forall ts1 ts2 e,
  lclosedTs ts1 ->
  lclosedTs ts2 ->
  opensAE (ts1 ++ ts2) e =
  opensAE ts2 (substsAE (map (shifts (length ts2)) (bnds (length ts1))) ts1 e).
Proof.
  induction ts1; intros ts2 e Hts1 Hts2; simpl.
    reflexivity.
    rewrite substAE_opensAE.
      rewrite IHts1. rewrite shifts_bnd. rewrite app_length.
      replace (length ts2 + length ts1) with (length ts1 + length ts2). 2: omega.
      fold (bnds (length ts1)). rewrite substAE_substsAE_bnd.
        reflexivity.
        omega.
        eapply lclosedTs_head. apply Hts1.
        eapply lclosedTs_tail. apply Hts1.
        eapply lclosedTs_tail. apply Hts1.
        apply Hts2.
        omega.
        eapply lclosedTs_head. apply Hts1.
        apply lclosedTs_concat. eapply lclosedTs_tail. apply Hts1. apply Hts2.
Qed.


(*************************************************************************)
(* Multiple opening on locally closed objects *)
(*************************************************************************)

Lemma opensAT_lclosed : forall t ts,
  lclosedT t ->
  opensAT ts t = t.
Proof.
  intros t ts H.
  unfold opensAT.
  induction ts.
    Case "nil".
      reflexivity.
    Case "t'::ts'".
      rewrite IHts.
      apply substAT_lclosed.
      apply H.
Qed.

Lemma opensAE_lclosed : forall e ts,
  lclosedE e ->
  opensAE ts e = e.
Proof.
  intros e ts H.
  unfold opensAE.
  induction ts.
    Case "nil".
      reflexivity.
    Case "t'::ts'".
      rewrite IHts.
      apply substAE_lclosed.
      apply H.
Qed.

Lemma opensXE_lclosed : forall e es,
  lclosedE e ->
  opensXE es e = e.
Proof.
  intros e es H.
  unfold opensXE.
  induction es.
    Case "nil".
      reflexivity.
    Case "e'::es'".
      rewrite IHes.
      apply substXE_lclosed.
      apply H.
Qed.


(*************************************************************************)
(* Multiple open and variables *)
(*************************************************************************)

Lemma opensAT_var_gt : forall ts n,
  length ts <= n ->
  opensAT ts n = n.
Proof.
  intros ts n.
  induction ts; intros Le; simpl opensAT.
    reflexivity.
    rewrite IHts; simpl length in Le.
      rewrite substAT_var_ne. reflexivity. unfold not. intros. inversion H. omega.
      omega.
Qed.

Lemma substATs_lclosed : forall (n : nat) t' ts,
  lclosedTs ts ->
  map (substAT n t') ts = ts.
Proof.
  intros n t'.
  induction ts; intro Hts; simpl map.
    reflexivity.
    f_equal.
      apply substAT_lclosed. eapply lclosedTs_head. apply Hts.
      apply IHts. eapply lclosedTs_tail. apply Hts.
Qed.

Lemma opensATs_bndTs : forall ts,
  lclosedTs ts ->
  map (opensAT ts) (bndTs (length ts)) = ts.
Proof.
  induction ts; intros Hts; simpl bndTs; simpl opensAT.
    apply map_id.
    simpl map.
      rewrite opensAT_var_gt. rewrite substAT_var.
        f_equal. rewrite <- map_map with (f := opensAT ts). unfold bndTs in IHts. unfold bnds in IHts. rewrite IHts.
          apply substATs_lclosed. eapply lclosedTs_tail. apply Hts.
          eapply lclosedTs_tail. apply Hts.
          omega.
Qed.


(*************************************************************************)
(* Multiple substitution and constructors *)
(*************************************************************************)

Lemma substsAT_ex : forall as' ts k t,
  substsAT as' ts (exT k t) = exT k (substsAT (map shift as') ts t).
Proof.
  induction as'; intros ts k t; simpl substsAT.
    reflexivity.
    destruct ts.
      reflexivity.
      rewrite IHas'. simpl substAT. reflexivity.
Qed.

Lemma substsAE_app : forall as' ts e1 e2,
  substsAE as' ts (appE e1 e2) = appE (substsAE as' ts e1) (substsAE as' ts e2).
Proof.
  induction as'; intros ts e1 e2; simpl substsAE.
    Case "ts = nil".
      reflexivity.
    Case "ts = t::ts'".
      destruct ts. reflexivity.
      rewrite IHas'. simpl substAE. reflexivity.
Qed.

Lemma substsAE_lam : forall as' ts t e,
  substsAE as' ts (lamE t e) = lamE (substsAT as' ts t) (substsAE as' ts e).
Proof.
  induction as'; intros ts t e; simpl substsAE.
    Case "ts = nil".
      reflexivity.
    Case "ts = t'::ts'".
      destruct ts. reflexivity.
      rewrite IHas'. simpl substsAE. reflexivity.
Qed.

Lemma substsAE_unpack : forall as' ts e1 e2,
  substsAE as' ts (unpackE e1 e2) = unpackE (substsAE as' ts e1) (substsAE (map shift as') ts e2).
Proof.
  induction as'; intros ts t e; simpl substsAE.
    Case "ts = nil".
      reflexivity.
    Case "ts = t'::ts'".
      destruct ts. reflexivity.
      rewrite IHas'. simpl substAE. reflexivity.
Qed.


(*************************************************************************)
(* Multiple open and constructors *)
(*************************************************************************)

Lemma opensAT_arr : forall ts t1 t2,
  opensAT ts (arrT t1 t2) = arrT (opensAT ts t1) (opensAT ts t2).
Proof.
  intros ts t1 t2.
  induction ts; simpl opensAT.
    Case "ts = nil".
      reflexivity.
    Case "ts = t::ts'".
      rewrite IHts. simpl substAT. reflexivity.
Qed.

Lemma opensAT_app : forall ts t1 t2,
  opensAT ts (appT t1 t2) = appT (opensAT ts t1) (opensAT ts t2).
Proof.
  intros ts t1 t2.
  induction ts; simpl opensAT.
    Case "ts = nil".
      reflexivity.
    Case "ts = t::ts'".
      rewrite IHts. simpl substAT. reflexivity.
Qed.

Lemma opensAT_ex : forall ts k t,
  opensAT ts (exT k t) = exT k (substsAT (map shift (bnds (length ts))) ts t).
Proof.
  induction ts; intros k t; simpl opensAT; simpl.
    reflexivity.
    rewrite IHts. simpl substAT. reflexivity.
Qed.


Lemma opensAE_var : forall ts x,
  opensAE ts (varE x) = varE x.
Proof.
  intros ts x.
  induction ts; simpl opensAE.
    Case "ts = nil".
      reflexivity.
    Case "ts = t'::ts'".
      rewrite IHts. simpl substAE. reflexivity.
Qed.

Lemma opensAE_lam : forall ts t e,
  opensAE ts (lamE t e) = lamE (opensAT ts t) (opensAE ts e).
Proof.
  intros ts t e.
  induction ts; simpl opensAE.
    Case "ts = nil".
      reflexivity.
    Case "ts = t'::ts'".
      rewrite IHts. simpl substAE. reflexivity.
Qed.

Lemma opensAE_app : forall ts e1 e2,
  opensAE ts (appE e1 e2) = appE (opensAE ts e1) (opensAE ts e2).
Proof.
  intros ts t e.
  induction ts; simpl opensAE.
    Case "ts = nil".
      reflexivity.
    Case "ts = t'::ts'".
      rewrite IHts. simpl substAE. reflexivity.
Qed.

Lemma opensAE_pack : forall ts t1 e2 t,
  opensAE ts (packE t1 e2 t) = packE (opensAT ts t1) (opensAE ts e2) (opensAT ts t).
Proof.
  intros ts t1 e2 t.
  induction ts; simpl opensAE; simpl opensAT.
    Case "ts = nil".
      reflexivity.
    Case "ts = t'::ts'".
      rewrite IHts. simpl substAE. reflexivity.
Qed.


(*************************************************************************)
(* Local closure under multiple substitution *)
(*************************************************************************)

Lemma lclosedT_substsAT : forall as' ts' t,
  lclosedTs ts' ->
  lclosedT t ->
  lclosedT (substsAT as' ts' t).
Proof.
  induction as'; intros ts' t Hts' Ht; simpl substsAT.
    apply Ht.
    destruct ts'.
      apply Ht.
      apply lclosedT_substAT.
        eapply lclosedTs_head. apply Hts'.
        apply IHas'. eapply lclosedTs_tail. apply Hts'. apply Ht.
Qed.

Corollary lclosedT_opensAT : forall ts' t,
  lclosedTs ts' ->
  lclosedT t ->
  lclosedT (opensAT ts' t).
Proof.
  intros ts' t Hts' Ht. rewrite unfold_opensAT. apply lclosedT_substsAT. apply Hts'. apply Ht.
Qed.

Lemma lclosedTs_substAT : forall a t' ts,
  lclosedT t' ->
  lclosedTs ts ->
  lclosedTs (map (substAT a t') ts).
Proof.
  intros a t'.
  induction ts; intros Ht' Hts; simpl map.
    apply lclosedTs_nil.
    apply lclosedTs_cons.
      apply lclosedT_substAT. apply Ht'. eapply lclosedTs_head. apply Hts.
      apply IHts. apply Ht'. eapply lclosedTs_tail. apply Hts.
Qed.

Corollary lclosedTs_openAT : forall t' ts,
  lclosedT t' ->
  lclosedTs ts ->
  lclosedTs (map (openAT t') ts).
Proof.
  intros t'.
  induction ts; intros Ht' Hts; simpl map.
    apply Hts.
    apply lclosedTs_cons.
      apply lclosedT_openAT. apply Ht'. eapply lclosedTs_head. apply Hts.
      apply IHts. apply Ht'.  eapply lclosedTs_tail. apply Hts.
Qed.

Lemma lclosedTs_substsAT : forall as' ts' ts,
  lclosedTs ts' ->
  lclosedTs ts ->
  lclosedTs (map (substsAT as' ts') ts).
Proof.
  intros as' ts'.
  induction ts; intros Hts' Hts; simpl map.
    apply lclosedTs_nil.
    apply lclosedTs_cons.
      apply lclosedT_substsAT. apply Hts'. eapply lclosedTs_head. apply Hts.
      apply IHts. apply Hts'. eapply lclosedTs_tail. apply Hts.
Qed.

Corollary lclosedTs_opensAT : forall ts' ts,
  lclosedTs ts' ->
  lclosedTs ts ->
  lclosedTs (map (opensAT ts') ts).
Proof.
  intros ts'.
  induction ts; intros Hts' Hts; simpl map.
    apply Hts.
    apply lclosedTs_cons.
      apply lclosedT_opensAT. apply Hts'. eapply lclosedTs_head. apply Hts.
      apply IHts. apply Hts'.  eapply lclosedTs_tail. apply Hts.
Qed.

Lemma notin_varsAT_substAT: forall a b t, a \notin {{b}} -> a \notin varsAT  t -> forall (n:nat), a \notin varsAT (substAT n b t).
Proof.
(* a shorter proof is this:
 intros a b.
 induction t; intros H0 H1 n; simpl substAT; simpl varsAT in * |- *; try apply notin_union_l; firstorder.
 destruct (eqOcc o n).
 apply H0.
 destruct o.
  apply H1.
  auto.
*)
 intros a b.
 induction t; intros H0 H1 n.
 Case "var".
 destruct o.
  SCase "free".
   simpl substAT.
   exact H1.
  SCase "bnd".
   unfold substAT.
   destruct (eqOcc n0 n).
    apply H0.
    apply H1.
 Case "unitT".
   apply H1.
 Case "arrT".
   simpl substAT.
   simpl varsAT in * |- *.
   apply notin_union_l.
   apply IHt1;auto.
   apply IHt2;auto.
 Case "timT".
   simpl substAT.
   simpl varsAT in * |- *.
   apply notin_union_l.
   apply IHt1;auto.
   apply IHt2;auto.
 Case "plusT".
   simpl substAT.
   simpl varsAT in * |- *.
   apply notin_union_l.
   apply IHt1;auto.
   apply IHt2;auto.
 Case "recT".
   simpl substAT.
   simpl varsAT in * |- *.
   apply IHt;auto.
 Case "alltT".
   simpl substAT.
   simpl varsAT in * |- *.
   apply IHt;auto.
 Case "exT".
   simpl substAT.
   simpl varsAT in * |- *.
   apply IHt;auto.
 Case "triv".
   apply H1.
 Case "lamT".
   simpl substAT.
   simpl varsAT in * |- *.
   apply IHt;auto.
 Case "appT".
   simpl substAT.
   simpl varsAT in * |- *.
   apply notin_union_l.
   apply IHt1;auto.
   apply IHt2;auto.
 Case "pairT".
   simpl substAT.
   simpl varsAT in * |- *.
   apply notin_union_l.
   apply IHt1;auto.
   apply IHt2;auto. 
 Case "fstT".
   simpl substAT.
   simpl varsAT in * |- *.
   apply IHt;auto.
 Case "sndT".
   simpl substAT.
   simpl varsAT in * |- *.
   apply IHt;auto.
Qed.

Corollary notin_varsAT_openAT : forall a b t, 
 a \notin {{b}} -> a \notin varsAT  t -> a \notin varsAT (openAT b t).
Proof.
  intros. unfold openAT. apply notin_varsAT_substAT; auto.
Qed.

Lemma notin_varsAT_opensAT : forall a' as' t L,
  fresh (L \u {{a'}}) (length as') as' -> a' \notin (varsAT t) -> a' \notin varsAT (opensAT (freeTs as') t).
Proof.
 intros a' as'.
 induction as' as [|b bs]; intros t L Frbbs Fra; try apply Frbs.
  simpl. 
 auto.
 intros.
 Case "var".
 simpl.
 rewrite substAT_opensAT.
 fold (freeTs bs).
 apply IHbs with (L := L).
 simpl fresh in Frbbs.
 destruct Frbbs; auto.
 apply notin_varsAT_substAT.
 simpl fresh in Frbbs.
 destruct Frbbs as [Frb Frbs].
 apply notin_singleton_swap.
 fresh_solve.
 auto.
 omega.
 auto.
 fold (freeTs bs).
 apply lclosedTs_vars.
Qed.
